package net.lamgc.scext.onedrive_transfer

import com.google.common.cache.CacheBuilder
import com.google.common.cache.CacheLoader
import com.google.common.util.concurrent.RateLimiter
import kotlinx.coroutines.delay
import kotlinx.coroutines.runBlocking
import mu.KotlinLogging
import org.telegram.telegrambots.abilitybots.api.bot.BaseAbilityBot
import org.telegram.telegrambots.meta.api.methods.send.SendMessage
import org.telegram.telegrambots.meta.api.methods.updatingmessages.DeleteMessage
import org.telegram.telegrambots.meta.api.methods.updatingmessages.EditMessageText
import java.util.*

@Suppress("UnstableApiUsage")
object DefaultOneDriveTransferCallback : OneDriveTransferCallback {

    private val logger = KotlinLogging.logger { }

    private val progressUpdateRateLimiterMap = WeakHashMap<BaseAbilityBot, RateLimiter>()
    private val progressUpdateUserRateLimiterMap = CacheBuilder.newBuilder()
        .expireAfterAccess(15, java.util.concurrent.TimeUnit.MINUTES)
        .softValues()
        .initialCapacity(12)
        .concurrencyLevel(4)
        .build(object : CacheLoader<BotUserRateLimitToken, RateLimiter>() {
            override fun load(key: BotUserRateLimitToken): RateLimiter {
                return RateLimiter.create(3.0)
            }
        })

    private fun BaseAbilityBot.rateLimiterPerBot(): RateLimiter {
        return progressUpdateRateLimiterMap[this] ?: RateLimiter.create(25.0).also {
            progressUpdateRateLimiterMap[this] = it
        }
    }

    private fun BaseAbilityBot.rateLimiterPerUser(userId: Long): RateLimiter {
        val token = BotUserRateLimitToken(this, userId)
        return progressUpdateUserRateLimiterMap[token]
    }

    private fun BaseAbilityBot.tryAcquireRateLimit(userId: Long): Boolean {
        return if (rateLimiterPerUser(userId).tryAcquire()) {
            rateLimiterPerBot().tryAcquire()
        } else {
            false
        }
    }

    override fun onTransferTaskCreated(task: OneDriveTransferTask) {
        if (task.extra["infoMessageId"] != null) {
            val deleteMessage = DeleteMessage.builder()
                .chatId(task.extra["chatId"].toString().toLong())
                .messageId(task.extra["infoMessageId"].toString().toInt())
                .build()

            try {
                task.bot.telegramClient.execute(deleteMessage)
            } catch (e: Exception) {
                logger.debug(e) { "删除旧状态消息时出错, 忽略该异常." }
            }
        }

        val msg = if (task.retryCount == 0) {
            """
                OneDrive 中转任务已创建
                正在排队中...
                文件名： ${task.document.fileName}
                -------------------------------------------------
                #Queuing
            """.trimIndent()
        } else {
            """
                OneDrive 中转任务已创建
                正在排队中...
                文件名： ${task.document.fileName}
                上次错误信息：${(task.extra["lastError"] as Exception?)?.message ?: "(没有错误信息)"}
                重试次数：${task.retryCount}
                -------------------------------------------------
                #Queuing
            """.trimIndent()
        }

        val message = task.bot.telegramClient.execute(
            SendMessage.builder()
                .text(msg)
                .chatId(task.extra["chatId"].toString().toLong())
                .replyToMessageId(task.extra["messageId"].toString().toInt())
                .build()
        )
        task.extra["infoMessageId"] = message.messageId
    }

    override fun onTransferTaskStart(progress: OneDriveTransferWorkerProgress) {
        val newMessage = EditMessageText.builder()
            .chatId(progress.currentTask.extra["chatId"].toString().toLong())
            .messageId(progress.currentTask.extra["infoMessageId"].toString().toInt())
            .text(
                """
                    OneDrive 中转任务开始执行
                    正在获取文件信息...(需要一些时间从 Telegram 服务器下载文件)
                    文件名： ${progress.currentTask.document.fileName}
                    -------------------------------------------------
                    #Starting
                    """.trimIndent()
            )
            .build().orSendMessage(progress.currentTask.bot, progress.currentTask.extra["messageId"].toString().toInt())
        if (newMessage != null) {
            progress.currentTask.extra["infoMessageId"] = newMessage.messageId
        }
    }

    override fun onGettingFileInfo(progress: OneDriveTransferWorkerProgress, retryCount: Int, maxRetryCount: Int) {
        val newMessage = EditMessageText.builder()
            .chatId(progress.currentTask.extra["chatId"].toString().toLong())
            .messageId(progress.currentTask.extra["infoMessageId"].toString().toInt())
            .text(
                """
                正在获取文件信息...
                文件名： ${progress.currentTask.document.fileName}
                重试次数：$retryCount / $maxRetryCount
                -------------------------------------------------
                重试并不等同于获取文件失败，由于 Telegram Bot API 需要下载文件后
                才会获取文件信息，因此需要重试以等待文件下载完成。
                
                #GettingFileInfo
            """.trimIndent()
            )
            .build().orSendMessage(progress.currentTask.bot, progress.currentTask.extra["messageId"].toString().toInt())
        if (newMessage != null) {
            progress.currentTask.extra["infoMessageId"] = newMessage.messageId
        }
    }

    override fun onGotFileInfo(
        progress: OneDriveTransferWorkerProgress,
        file: org.telegram.telegrambots.meta.api.objects.File
    ) {
        val newMessage = EditMessageText.builder()
            .chatId(progress.currentTask.extra["chatId"].toString().toLong())
            .messageId(progress.currentTask.extra["infoMessageId"].toString().toInt())
            .text(
                """
                已获取文件信息，正在创建 OneDrive 上传会话...
                文件名： ${progress.currentTask.document.fileName}
                文件大小：${file.fileSize / 1024} KB
                -------------------------------------------------
                #UploadStarting
                """.trimIndent()
            )
            .build().orSendMessage(progress.currentTask.bot, progress.currentTask.extra["messageId"].toString().toInt())
        if (newMessage != null) {
            progress.currentTask.extra["infoMessageId"] = newMessage.messageId
        }
    }

    override fun onUploadProgress(progress: OneDriveTransferWorkerProgress) {
        if (!progress.currentTask.bot.tryAcquireRateLimit(progress.currentTask.tgUserId)) {
            return
        }
        val newMessage = EditMessageText.builder()
            .chatId(progress.currentTask.extra["chatId"].toString().toLong())
            .messageId(progress.currentTask.extra["infoMessageId"].toString().toInt())
            .text(
                """
                OneDrive 中转任务执行中
                文件名： ${progress.currentTask.document.fileName}
                进度：${String.format("%.3f", progress.progress.get() * 100)}%
                -------------------------------------------------
                #Processing 
            """.trimIndent()
            )
            .build().orSendMessage(progress.currentTask.bot, progress.currentTask.extra["messageId"].toString().toInt())
        if (newMessage != null) {
            progress.currentTask.extra["infoMessageId"] = newMessage.messageId
        }
    }

    override fun onTransferFailure(task: OneDriveTransferTask, progress: OneDriveTransferWorkerProgress) {
        val newMessage = EditMessageText.builder()
            .chatId(task.extra["chatId"].toString().toLong())
            .messageId(task.extra["infoMessageId"].toString().toInt())
            .text(
                """
                OneDrive 中转任务执行失败
                文件名： ${task.document.fileName}
                错误信息：${progress.exception?.message}
                重试次数：${task.retryCount}
                
                任务将会追加至队列尾部进行重试。
                -------------------------------------------------
                #Failure #Error
            """.trimIndent()
            )
            .build().orSendMessage(progress.currentTask.bot, progress.currentTask.extra["messageId"].toString().toInt())
        if (newMessage != null) {
            progress.currentTask.extra["infoMessageId"] = newMessage.messageId
            progress.currentTask.extra["lastError"] = progress.exception
        }
        if (task.retryCount < 5) {
            runBlocking {
                delay(10000)
                task.retryCount++
                OneDriveTransferCenter.submitUploadTask(task)
            }
        }
    }

    override fun onTransferSuccess(task: OneDriveTransferTask, progress: OneDriveTransferWorkerProgress) {
        val newMessage = EditMessageText.builder()
            .chatId(task.extra["chatId"].toString().toLong())
            .messageId(task.extra["infoMessageId"].toString().toInt())
            .text(
                """
                OneDrive 中转任务执行成功
                文件名： ${task.document.fileName}
                OneDrive 文件路径：${progress.driveItem?.webUrl}
                ------------------------------------------------
                #Success
            """.trimIndent()
            )
            .build().orSendMessage(progress.currentTask.bot, progress.currentTask.extra["messageId"].toString().toInt())
        if (newMessage != null) {
            progress.currentTask.extra["infoMessageId"] = newMessage.messageId
        }
    }

    override fun onTransferCancelled(task: OneDriveTransferTask, progress: OneDriveTransferWorkerProgress?) {
        val newMessage = EditMessageText.builder()
            .chatId(task.extra["chatId"].toString().toLong())
            .messageId(task.extra["infoMessageId"].toString().toInt())
            .text(
                """
                OneDrive 中转任务已取消
                文件名： ${task.document.fileName}
                -------------------------------------------------
                #Cancelled
            """.trimIndent()
            )
            .build().orSendMessage(task.bot, task.extra["messageId"].toString().toInt())
        if (newMessage != null) {
            task.extra["infoMessageId"] = newMessage.messageId
        }
    }

}

private data class BotUserRateLimitToken(
    val bot: BaseAbilityBot,
    val userId: Long
)
