package net.lamgc.scext.onedrive_transfer

import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.module.kotlin.registerKotlinModule
import com.google.common.cache.CacheBuilder
import com.microsoft.aad.msal4j.MsalInteractionRequiredException
import mu.KotlinLogging
import org.jetbrains.exposed.sql.Database
import org.telegram.telegrambots.abilitybots.api.bot.BaseAbilityBot
import org.telegram.telegrambots.abilitybots.api.objects.Ability
import org.telegram.telegrambots.abilitybots.api.objects.Locality
import org.telegram.telegrambots.abilitybots.api.objects.Privacy
import org.telegram.telegrambots.abilitybots.api.objects.Reply
import org.telegram.telegrambots.abilitybots.api.util.AbilityExtension
import org.telegram.telegrambots.meta.api.methods.send.SendMessage
import org.telegram.telegrambots.meta.api.methods.updatingmessages.DeleteMessage
import org.telegram.telegrambots.meta.api.methods.updatingmessages.EditMessageText
import org.telegram.telegrambots.meta.api.objects.replykeyboard.InlineKeyboardMarkup
import org.telegram.telegrambots.meta.api.objects.replykeyboard.ReplyKeyboardMarkup
import org.telegram.telegrambots.meta.api.objects.replykeyboard.ReplyKeyboardRemove
import org.telegram.telegrambots.meta.api.objects.replykeyboard.buttons.InlineKeyboardButton
import org.telegram.telegrambots.meta.api.objects.replykeyboard.buttons.InlineKeyboardRow
import org.telegram.telegrambots.meta.api.objects.replykeyboard.buttons.KeyboardRow
import java.io.File
import java.net.MalformedURLException
import java.net.URL
import java.util.concurrent.TimeUnit

class OneDriveTransferExtension(val bot: BaseAbilityBot, val dataFolder: File) : AbilityExtension {

    private val logger = KotlinLogging.logger {  }

    private val config: ExtensionConfig
    private val onedriveService: OneDriveTransferService
    private val buttonTokenCache = CacheBuilder.newBuilder()
        .expireAfterWrite(5, TimeUnit.MINUTES)
        .build<String, String>()

    private val actionCache = CacheBuilder.newBuilder()
        .expireAfterWrite(5, TimeUnit.MINUTES)
        .build<Long, String>()

    init {
        config = loadConfiguration()
        val db = Database.connect("jdbc:sqlite:${File(dataFolder, "./data.db").canonicalPath}", "org.sqlite.JDBC")
        onedriveService = OneDriveTransferService(bot, config, db)
    }

    fun loadConfiguration(): ExtensionConfig {
        val configFile = File(this.dataFolder, "config.json")
        val objectMapper = ObjectMapper().registerKotlinModule()
        if (!configFile.exists()) {
            configFile.createNewFile()
            objectMapper.writeValue(configFile, ExtensionConfig())
            return ExtensionConfig()
        }
        return objectMapper.readValue(configFile, ExtensionConfig::class.java)
    }

    fun startToUsage(): Ability = Ability.builder()
        .named("start")
        .info("查看 OneDrive Transfer 的帮助信息.")
        .privacy(Privacy.PUBLIC)
        .locality(Locality.USER)
        .action {
            it.bot().silent.send(
                """
                OneDrive Transfer 是一个 Telegram 机器人，
                可以将 Telegram 中的文件上传到 OneDrive 中转。
                
                /login          登录 OneDrive 账户
                /logout         登出 OneDrive 账户。
                /my             查看当前 OneDrive 账户信息。
                /select_drive   选择 OneDrive 驱动器。
                /set_path       设置 OneDrive 中转路径。
                /cancel         取消当前操作（仅限 set_path 命令）。
                /start          查看帮助信息。
                
                --------------------------------------------
                基本用法：
                1. 先使用 /login 命令，按照提示登录要存储文件的 OneDrive 账户。
                2. 登录完成后执行 /select_drive 选择 OneDrive 驱动器；
                3. 选好后直接转发文件给机器人就好，机器人会自动上传到 OneDrive 中转。
                
                文件默认保存在 "Telegram Files/"，如果需要修改可以使用 /set_path 命令。
            """.trimIndent(), it.chatId()
            )
        }
        .build()

    fun loginOneDrive(): Ability = Ability
        .builder()
        .named("login")
        .info("登录 OneDrive 账户.")
        .locality(Locality.USER)
        .privacy(Privacy.PUBLIC)
        .action { ctx ->
            val url = onedriveService.createLoginUrl(ctx.chatId())
            actionCache.put(ctx.chatId(), "login")
            ctx.bot().silent.send(
                """
                请使用以下链接进行登录：
                $url
                
                -------------------------------------------
                登录成功后会显示无法访问，是正常情况，请将地址栏的链接发回给机器人。
            """.trimIndent(), ctx.chatId())
        }
        .enableStats()
        .reply(Reply.of(
            {bot, upd ->
                try {
                    val account = onedriveService.updateAccount(upd.message.chat.id, URL(upd.message.text.trim()))
                    actionCache.invalidate(upd.message.chatId)
                    bot.silent.send(
                        """
                        登录成功！
                        Microsoft 账号：${account.userName}
                        请使用 /select_drive 选择 OneDrive 驱动器以及设置上传路径。
                    """.trimIndent(), upd.message.chatId)
                } catch (e: MsalInteractionRequiredException) {
                    if (e.errorCode() == "AADSTS54005") {
                        bot.silent.send("登录失败，登录链接已过期，请重新登录。", upd.message.chatId)
                    } else {
                        bot.silent.send("登录失败，错误代码：${e.errorCode()}", upd.message.chatId)
                    }
                    actionCache.invalidate(upd.message.chatId)
                } catch (e: MalformedURLException) {
                    bot.silent.send("链接格式错误，请发送正确的登录链接。", upd.message.chatId)
                } catch (e: Exception) {
                    logger.error(e) { "处理 Oauth2 令牌时发生错误." }
                    bot.silent.send("处理 Oauth2 令牌时发生错误，请稍后重试。", upd.message.chatId)
                    actionCache.invalidate(upd.message.chatId)
                }
            },
            { upd -> upd.hasMessage() && upd.message.hasText() },
            { upd -> actionCache.getIfPresent(upd.message.chatId) == "login" }
        ))
        .build()

    fun status(): Ability = Ability.builder()
        .named("my")
        .info("查看当前 OneDrive 设定信息.")
        .locality(Locality.USER)
        .privacy(Privacy.PUBLIC)
        .action {
            val account = onedriveService.accountManager.getAccountByTgUserId(it.chatId())
            if (account == null) {
                it.bot().silent.send("当前账户未登录 OneDrive.", it.chatId())
                return@action
            }
            it.bot().silent.send(
                """
                当前账户已登录 OneDrive.
                Microsoft 账号： ${account.userName}
            """.trimIndent(), it.chatId())
        }
        .build()

    fun selectDrive(): Ability = Ability.builder()
        .named("select_drive")
        .info("选择 OneDrive 驱动器.")
        .locality(Locality.USER)
        .privacy(Privacy.PUBLIC)
        .action {
            val currentDrive = try {
                onedriveService.getCurrentDrive(it.chatId())
            } catch (e: OneDriveNotLoginException) {
                it.bot().silent.send("当前账户没有登录 OneDrive.", it.chatId())
                return@action
            } catch (e: Exception) {
                null
            }
            val drives = onedriveService.listDriversByUserId(it.chatId())
            if (drives.isEmpty()) {
                it.bot().silent.send("当前账户没有 OneDrive 驱动器.", it.chatId())
                return@action
            }

            val msgContent = """
                当前账户已登录 OneDrive.
                Microsoft 账号： ${onedriveService.accountManager.getAccountByTgUserId(it.chatId())?.userName}
                当前正在使用的驱动器为：[${currentDrive?.driveType ?: "None" }] ${currentDrive?.name ?: "无"}
                """.trimIndent()

            SendMessage.builder()
                .disableWebPagePreview(true)
                .chatId(it.chatId().toString())
                .text(msgContent)
                .replyMarkup(InlineKeyboardMarkup.builder().apply {
                    for (drive in drives) {
                        keyboardRow(
                            InlineKeyboardRow(
                                InlineKeyboardButton.builder()
                                    .text("[${drive.driveType}] ${drive.name}")
                                    .callbackData(
                                        buttonTokenCache.putToken(
                                            "${drive.id}",
                                            tokenPrefix = "select_drive:"
                                        )
                                    )
                                    .build()
                            )
                        )
                    }
                }.build())
                .build().let { msg ->
                    it.bot().silent.execute(msg)
                }
        }
        .reply({ bot, upd ->
            val driveId = buttonTokenCache.getIfPresent(upd.callbackQuery.data.substringAfter(':'))
            if (driveId == null) {
                bot.silent.send("无效的驱动器 ID.", upd.callbackQuery.message.chatId)
                return@reply
            }
            val drive = try {
                onedriveService.listDriversByUserId(upd.callbackQuery.from.id).firstOrNull { it.id == driveId }
            } catch (e: OneDriveNotLoginException) {
                bot.silent.send("当前账户没有登录 OneDrive.", upd.callbackQuery.message.chatId)
                return@reply
            }
            if (drive == null) {
                bot.silent.send("无效的驱动器 ID.", upd.callbackQuery.message.chatId)
                return@reply
            }
            onedriveService.setDrive(upd.callbackQuery.from.id, driveId)
            val editMessageText = EditMessageText.builder()
                .chatId(upd.callbackQuery.message.chatId)
                .messageId(upd.callbackQuery.message.messageId)
                .replyMarkup(
                    InlineKeyboardMarkup.builder()
                        .clearKeyboard()
                        .build()
                )
                .text(
                    """
                    已选择驱动器：[${drive.driveType}] ${drive.name}
                    请使用 /set_path 设置上传路径。
                    """.trimIndent()
                )
                .build()
            bot.silent.execute(editMessageText)
        }, { upd ->
            upd.hasCallbackQuery() && upd.callbackQuery.data != null && upd.callbackQuery.data.startsWith("select_drive:")
        })
        .build()

    fun setPath(): Ability = Ability.builder()
        .named("set_path")
        .info("设置 OneDrive 中转路径.")
        .locality(Locality.USER)
        .privacy(Privacy.PUBLIC)
        .action {
            val currentDrive = try {
                onedriveService.getCurrentDrive(it.chatId())
            } catch (e: OneDriveNotLoginException) {
                it.bot().silent.send("当前账户没有登录 OneDrive.", it.chatId())
                return@action
            }
            if (currentDrive == null) {
                it.bot().silent.send("当前账户没有选择 OneDrive 驱动器.", it.chatId())
                return@action
            }
            val transferSetting = onedriveService.accountManager.getTransferSetting(it.chatId())
            if (transferSetting == null) {
                it.bot().silent.send("当前账户没有登录 OneDrive.", it.chatId())
                return@action
            }
            val msgContent = """
                当前账户已登录 OneDrive.
                Microsoft 账号： ${transferSetting.userName}
                当前正在使用的驱动器为：[${currentDrive.driveType}] ${currentDrive.name}
                当前上传路径为：${transferSetting.storagePath}
                -------------------------------------------------
                请发送新的上传路径，或者发送 /cancel 取消设置。
                """.trimIndent()
            actionCache.put(it.chatId(), "set_path")
            SendMessage.builder()
                .disableWebPagePreview(true)
                .chatId(it.chatId().toString())
                .text(msgContent)
                .replyMarkup(ReplyKeyboardMarkup.builder().apply {
                    this.oneTimeKeyboard(true)
                    this.selective(false)
                    this.inputFieldPlaceholder("OneDrive 上传路径")
                    this.keyboardRow(KeyboardRow().apply {
                        add("Telegram Files/")
                    })
                    this.resizeKeyboard(true)
                    this.isPersistent(false)
                }.build())
                .build().let { msg ->
                    it.bot().silent.execute(msg)
                }
        }
        .reply({ bot, upd ->
            val path = upd.message.text.trim()
            if (path == "/cancel") {
                actionCache.invalidate(upd.message.chatId)
                bot.silent.execute(
                    SendMessage.builder()
                    .chatId(upd.message.chatId.toString())
                    .text("已取消设置.")
                    .replyMarkup(ReplyKeyboardRemove.builder()
                        .removeKeyboard(true)
                        .build())
                    .build())
                return@reply
            }
            val transferSetting = onedriveService.accountManager.getTransferSetting(upd.message.chatId)
            if (transferSetting == null) {
                actionCache.invalidate(upd.message.chatId)
                bot.silent.execute(
                    SendMessage.builder()
                    .chatId(upd.message.chatId.toString())
                    .text("当前账户没有登录 OneDrive.")
                    .replyMarkup(ReplyKeyboardRemove.builder()
                        .removeKeyboard(true)
                        .build())
                    .build())
                return@reply
            }
            onedriveService.accountManager.doSomething {
                transferSetting.storagePath = path
            }
            actionCache.invalidate(upd.message.chatId)
            bot.silent.execute(
                DeleteMessage.builder()
                .chatId(upd.message.chatId.toString())
                .messageId(upd.message.messageId)
                .build())
            bot.silent.execute(
                SendMessage.builder()
                .chatId(upd.message.chatId.toString())
                .text("已设置上传路径为：$path")
                .replyMarkup(ReplyKeyboardRemove.builder()
                    .removeKeyboard(true)
                    .build())
                .build())
        }, { upd ->
            upd.hasMessage() && upd.message.hasText() && actionCache.getIfPresent(upd.message.chatId) == "set_path"
        })
        .build()

    fun logout(): Ability = Ability.builder()
        .named("logout")
        .info("登出 OneDrive 账户.")
        .locality(Locality.USER)
        .privacy(Privacy.PUBLIC)
        .action {
            val transferSetting = onedriveService.accountManager.getTransferSetting(it.chatId())
            if (transferSetting == null) {
                it.bot().silent.send("当前账户没有登录 OneDrive.", it.chatId())
                return@action
            }
            onedriveService.accountManager.doSomething {
                transferSetting.delete()
            }
            it.bot().silent.send("已登出 OneDrive.", it.chatId())
        }
        .build()

    fun uploadDocument(): Reply = Reply.of(
        { bot, upd ->
            val document = upd.message.document
            if (document == null) {
                bot.silent.send("请发送文件.", upd.message.chatId)
                return@of
            }
            if (document.fileSize > config.maxFileSize) {
                bot.silent.send("文件大小超过限制.", upd.message.chatId)
                return@of
            }
            onedriveService.submitUploadDocumentTask(upd.message.chatId, upd.message.messageId, document)
        },
        { upd -> upd.hasMessage() && upd.message.hasDocument() },
        { upd ->
            try {
                val transferSetting = onedriveService.accountManager.getTransferSetting(upd.message.chatId)
                transferSetting != null
            } catch (e: OneDriveNotLoginException) {
                false
            }
        }
    )

    fun adminStatus(): Ability = Ability.builder()
        .named("status")
        .info("查看当前中转状态.(仅限管理员)")
        .privacy(Privacy.CREATOR)
        .locality(Locality.USER)
        .action {
            val processingTasks = OneDriveTransferCenter.getProcessingTasks()
            val msg = StringBuilder(
                """
                当前排队任务数：${OneDriveTransferCenter.getQueueingTaskCount()}
                当前处理任务数：${processingTasks.size} / ${OneDriveTransferCenter.getWorkerCount()}
                --------------------------------------------------------------
                """.trimIndent()
            ).append('\n')
            for (workerId in 0..<OneDriveTransferCenter.getWorkerCount()) {
                val task = processingTasks[workerId]
                if (task == null) {
                    msg.append("Worker $workerId: 空闲\n")
                } else {
                    msg.append(
                        "Worker $workerId: [${task.status}] ${
                            String.format(
                                "%.3f",
                                task.progress.get() * 100.0
                            )
                        }% ${task.currentTask.document.fileSize} Bytes\n"
                    )
                }
            }
            it.bot().silent.send(msg.toString(), it.chatId())
        }
        .build()

    fun cancelTask(): Ability = Ability.builder()
        .named("cancel_task")
        .info("取消指定的任务（回复对应的文件或状态消息）")
        .privacy(Privacy.PUBLIC)
        .locality(Locality.USER)
        .action {
            if (!it.update().message.isReply) {
                it.bot().silent.send("请回复要取消的消息.", it.chatId())
                return@action
            }

            val replyMessage = it.update().message.replyToMessage
            val result = onedriveService.cancelTask(it.chatId(), replyMessage.messageId)
            val msg = if (result) {
                "已取消任务."
            } else {
                "任务已完成，无法取消."
            }
            val sendMessage = SendMessage.builder()
                .chatId(it.chatId().toString())
                .text(msg)
                .replyToMessageId(it.update().message.messageId)
                .build()
            it.bot().silent.execute(sendMessage)
        }
        .build()

    private fun Ability.AbilityBuilder.named(name: String): Ability.AbilityBuilder {
        return if (config.useCommandPrefix) {
            name("odt_$name")
        } else {
            name(name)
        }
    }

}
